c*********************************************************************
c*********************************************************************
      Program main 
c*********************************************************************
c*********************************************************************
c *******************************************************************************
c *                   GDEP. This version 2.0 Oct 31, 2008                       *
c * This FORTRAN program is used to compute transition dependency:              *
c * A gene-gene interaction measure for times series microarray data            *
c * Reference: Xin Gao, Daniel Q. Pu, Peter X-K. Song                           *
c * EURASIP Journal on Bioinformatics and System Biology(2008)                  *
c * This program is written by Daniel Q. Pu, Xin Gao, Peter X-K. Song           *
c * Comments and suggestions are welcome                                        *
c * Please cite the reference source for any usage of this program              *
c * This program also cites certain subroutine from RANLIB                      *
c * Library of Fortran Routines for Random Number Generation                    *
c * The corresponding author and references have been cited in the program      *
c *                                                                             *
c * XIN GAO, Department of Mathematics and Statistics, York University          *
c * xingao@mathstat.yorku.ca                                                    *
c * DANIEL Q. PU                                                                *
c * Fraud Loss Prevention&Investigation, Fraud Control, CIBC, Toronto, CA       *
c * qiang.pu@cibc.com                                                           *
c * Peter X-K. Song                                                             *
c * Department of Biostatistics, University of Michigan School of Public Health *
c * pxsong@umich.edu                                                            *
c *                                                                             *
c * Input:                                                                      *
c * "realdata.txt" is the dataset with format(T*REP,Ngene);                     *
c * "para1.txt" includes the mean and std with format(2, Ngene);                *
c * "para2.txt" includes the mean and std with format(2, Ngene);                *
c *                                                                             *
c * Output: "output.txt" and "pvalue.txt"                                       *
c * "output.txt" includes the following information:                            *
c * nullstat: includes the Boot-strap value under null based on proposed method;*
c * nullcorr: includes the Boot-strap value under null with correlation method; *
c * "pvalue.txt" includes the following information:                            *
c * observerd statistic with proposed method;                                   *    
c * pvalue of each pair based on our proposed method;                           *
c * observerd statistic with correlation method;                                *    
c * pvalue of each pair based on correlation method;                            *
c *******************************************************************************
c
c


      Integer T,REP,B,i,j,u,v,N,tt,bb,batch,Ngene
      Integer NT,NREP,NB,NNgene
      Double Precision corr,ssum,correlation
      Double Precision chivec,nullmat(4,4),pivec(4)
      Double Precision TMAT(4,4),ZSTAT

      INTEGER IR(4),C(4),morcode(4,2)
      INTEGER cornum, resnum,zstatnum
      REAL P(4), snorm
      INTEGER uu,vv,ss,indm1,indm2, uu1,vv1
      Double Precision R(1), mm, std,d1,d2,temp3(4)
      CHARACTER FLAG

    
      Double Precision, allocatable ::  res(:,:), corres(:,:)
      Double Precision, allocatable ::  zstatres(:,:)
      Double Precision, allocatable ::  nullzstat(:,:,:)
      Double Precision, allocatable ::  nullcorr(:,:,:)
      Double Precision, allocatable ::  nullstat(:,:,:)
      Double Precision, allocatable ::  yex(:,:,:),qvec(:,:,:)
      INTEGER, allocatable ::  st(:,:),bst(:,:,:)
      Double Precision, allocatable ::  temp1(:),temp2(:)
      Double Precision, allocatable ::  modtcell44(:,:)
      Double Precision, allocatable ::  para1(:,:),para2(:,:)
      REAL, allocatable ::  corpval(:,:),newpval(:,:)


      WRITE(*,*) 'Probability density is gene index or time index?'     
      WRITE(*,*) '(G: gene index; T: time index)'
      READ(*,*) FLAG

  71  WRITE(*,*) 'Input time points:'
      READ(*,*) T
      IF (T .GT. 50) THEN
      WRITE(*,*) 'Input error: time points are suggested less than 50!'
              go to 71
      ENDIF        

  72  WRITE(*,*) 'Input the number of replicates:'
      READ(*,*) REP
      IF (REP .GT. 50) THEN
              WRITE(*,*) 'Input error: The number of replicates 
     +are suggested than 50!'
              go to 72
      ENDIF        

  73  WRITE(*,*) 'Input the number of boot strap:'
      READ(*,*) B
      IF (B .GT. 100) THEN
              WRITE(*,*) 'Input error: The number of boot strap are 
     +suggested than 100!'
              go to 73
      ENDIF        

   74 WRITE(*,*) 'Input the number of genes:'
      READ(*,*) Ngene
      IF (Ngene .GT. 60) THEN
        WRITE(*,*) 'Input error: the number of genes are suggested 
     +than 100!'
              go to 74
      ENDIF        
      WRITE(*,*) 'The program is running, please wait...'

      allocate(res(Ngene,Ngene))
      allocate(corres(Ngene,Ngene))
      allocate(zstatres(Ngene,Ngene))
      allocate(nullzstat(Ngene,Ngene,B))
      allocate(nullcorr(Ngene,Ngene,B))
      allocate(nullstat(Ngene,Ngene,B))
      allocate(yex(REP,T,2))
      allocate(temp1(REP*T))
      allocate(temp2(REP*T))
      allocate(modtcell44(T*REP,Ngene))
      allocate(st(REP,T))
      allocate(bst(REP,T,2))
      allocate(para1(2,Ngene))
      allocate(para2(2,Ngene))
      allocate(corpval(Ngene,Ngene))
      allocate(newpval(Ngene,Ngene))
      allocate(qvec(REP,T,4))

      open(unit=7, file='pvalue.txt',status='new')
      open(unit=8, file='output.txt',status='new')
      write (7,333)
      write (8,444)

      DATA morcode/0,0,1,1,0,1,0,1/
      DATA C/1,2,3,4/  


      open(unit=91,file='realdata.txt',status='old')
       do 81 i=1,T*REP
       read(91,*) modtcell44(i,1:Ngene)
   81  continue
       close(91)


      open(unit=93,file='para1.txt',status='old')
       do 83 i=1,2
       read(93,*) para1(i,1:Ngene)
   83  continue
       close(93)

      open(unit=94,file='para2.txt',status='old')
       do 84 i=1,2
       read(94,*) para2(i,1:Ngene)
   84  continue
       close(94)

       DO 100 uu=1, (Ngene-1)
          DO 121 vv=(uu+1), Ngene

       i=1
       DO 101 tt=1, T
           DO 102 N=1, REP

           yex(N,tt,1)=modtcell44(i,uu)
           yex(N,tt,2)=modtcell44(i,vv)  
           i=i+1

  102 continue
  101 continue

! Correlation between yex[1:REP,1:T,1] and yex[1:REP,1:T,2]

      do 110 N=1, REP
      temp1(((N-1)*T+1):(N*T))=yex(N,1:T,1)
      temp2(((N-1)*T+1):(N*T))=yex(N,1:T,2)
  110 continue  
      correlation=corr(REP*T,temp1,temp2)    

! Z statistic under null          
      ZSTAT=log((1+correlation)/(1-correlation))/(REP*T-1)   

      DO 410 N=1, REP
         DO 411 tt=1, T
            DO 412 ss=1,4    
            indm1=morcode(ss,1)
            indm2=morcode(ss,2)

            IF (indm1.EQ.0) THEN
            mm=para1(1,uu) 
            std=para1(2,uu)
            d1=dnorm(yex(N,tt,1),mm,std)
            ELSE
            mm=para2(1,uu) 
            std=para2(2,uu)
            d1=dnorm(yex(N,tt,1),mm,std)
            ENDIF  

            IF (indm2.EQ.0) THEN
            mm=para1(1,vv) 
            std=para1(2,vv)
            d2=dnorm(yex(N,tt,2),mm,std)
            ELSE
            mm=para2(1,vv) 
            std=para2(2,vv)
            d2=dnorm(yex(N,tt,2),mm,std)
            ENDIF  

            IF (FLAG .EQ. 'T') THEN
            IF (indm1.EQ.0) THEN
            mm=para1(1,tt) 
            std=para1(2,tt)
            d1=dnorm(yex(N,tt,1),mm,std)
            ELSE
            mm=para2(1,tt) 
            std=para2(2,tt)
            d1=dnorm(yex(N,tt,1),mm,std)
            ENDIF  

            IF (indm2.EQ.0) THEN
            mm=para1(1,tt) 
            std=para1(2,tt)
            d2=dnorm(yex(N,tt,2),mm,std)
            ELSE
            mm=para2(1,tt) 
            std=para2(2,tt)
            d2=dnorm(yex(N,tt,2),mm,std)
            ENDIF  
            ENDIF

        qvec(N,tt,ss)=d1*d2

  412 CONTINUE
  411 CONTINUE
  410 CONTINUE

      CALL chivecfun (REP,T,qvec,chivec,nullmat,pivec)


      corres(uu,vv)=abs(correlation)
      res(uu,vv)=chivec
      zstatres(uu,vv)=abs(ZSTAT)

      DO 200 bb=1, B
 
      CALL normalvec(4,(pivec(1:4)+(1.0D-6)),temp3(1:4))
      pivec=temp3
      CALL normalization(4,4,(nullmat+(1.0D-6)),TMAT)
      
       DO 210 N=1, rep
       tt=1
       P(1:4)=REAL(pivec(1:4))
       CALL MTN (1,P,IR,4) 
       st(N,tt)=dot_product(IR,C)   

! joint binary state
       bst(N, tt, 1:2)=morcode(st(N, tt), 1:2)

       DO 250 tt=2, T
       P(1:4)=REAL(TMAT(st(N,tt-1),1:4)) 
       CALL MTN (1,P,IR,4) 
       st(N,tt)=dot_product(IR,C)      
       bst(N, tt, 1:2)=morcode(st(N, tt), 1:2)
  250  continue
       
       DO 230 tt=1, T
! Gene index
       IF (FLAG .EQ. 'G') THEN
       IF (bst(N,tt,1).EQ.0) THEN
          R=snorm()
          mm=para1(1,uu)
          std=para1(2,uu)
          yex(N,tt,1)=R(1)*std+mm
          ELSE   
          mm=para2(1,uu)
          std=para2(2,uu)
          yex(N,tt,1)=R(1)*std+mm
       ENDIF   

       IF (bst(N,tt,2).EQ.0) THEN
          R=snorm()
          mm=para1(1,vv)
          std=para1(2,vv)
          yex(N,tt,2)=R(1)*std+mm
          ELSE   
          mm=para2(1,vv)
          std=para2(2,vv)
          yex(N,tt,2)=R(1)*std+mm
       ENDIF   
       ENDIF 

! Time index
       IF (FLAG .EQ. 'T') THEN
       IF (bst(N,tt,1).EQ.0) THEN
          R=snorm()
          mm=para1(1,tt)
          std=para1(2,tt)
          yex(N,tt,1)=R(1)*std+mm
          ELSE   
          mm=para2(1,tt)
          std=para2(2,tt)
          yex(N,tt,1)=R(1)*std+mm
       ENDIF   

       IF (bst(N,tt,2).EQ.0) THEN
          R=snorm()
          mm=para1(1,tt)
          std=para1(2,tt)
          yex(N,tt,2)=R(1)*std+mm
          ELSE   
          mm=para2(1,tt)
          std=para2(2,tt)
          yex(N,tt,2)=R(1)*std+mm
       ENDIF   
       ENDIF

  230 continue  
  210 continue

      do 300 N=1, REP
      temp1(((N-1)*T+1):(N*T))=yex(N,1:T,1)
      temp2(((N-1)*T+1):(N*T))=yex(N,1:T,2)
  300 continue  
      correlation=corr(REP*T,temp1,temp2)    


! Z statistic under null          
      ZSTAT=log((1+correlation)/(1-correlation))/(REP*T-1)   

      DO 510 N=1, REP
         DO 511 tt=1, T
            DO 512 ss=1,4    
            indm1=morcode(ss,1)
            indm2=morcode(ss,2)
! Gene index            
            IF (FLAG .EQ. 'G') THEN
            IF (indm1.EQ.0) THEN
            mm=para1(1,uu) 
            std=para1(2,uu)
            d1=dnorm(yex(N,tt,1),mm,std)
            ELSE
            mm=para2(1,uu) 
            std=para2(2,uu)
            d1=dnorm(yex(N,tt,1),mm,std)
            ENDIF  

            IF (indm2.EQ.0) THEN
            mm=para1(1,vv) 
            std=para1(2,vv)
            d2=dnorm(yex(N,tt,2),mm,std)
            ELSE
            mm=para2(1,vv) 
            std=para2(2,vv)
            d2=dnorm(yex(N,tt,2),mm,std)
            ENDIF  
            ENDIF  
! Time index
            IF (FLAG .EQ. 'T') THEN
            IF (indm1.EQ.0) THEN
            mm=para1(1,tt) 
            std=para1(2,tt)
            d1=dnorm(yex(N,tt,1),mm,std)
            ELSE
            mm=para2(1,tt) 
            std=para2(2,tt)
            d1=dnorm(yex(N,tt,1),mm,std)
            ENDIF  

            IF (indm2.EQ.0) THEN
            mm=para1(1,tt) 
            std=para1(2,tt)
            d2=dnorm(yex(N,tt,2),mm,std)
            ELSE
            mm=para2(1,tt) 
            std=para2(2,tt)
            d2=dnorm(yex(N,tt,2),mm,std)
            ENDIF  
            ENDIF

        qvec(N,tt,ss)=d1*d2

  512 CONTINUE
  511 CONTINUE
  510 CONTINUE


      CALL chivecfun (REP,T,qvec,chivec,nullmat,pivec)
      nullcorr(uu,vv,bb)=abs(correlation)
      nullstat(uu,vv,bb)=chivec
      nullzstat(uu,vv,bb)=abs(ZSTAT)

       write (8,555) uu,vv,bb,nullstat(uu,vv,bb),nullcorr(uu,vv,bb)
  200 continue       
  121  CONTINUE     
  100  CONTINUE     

       DO 900 uu=1, (Ngene-1)
          DO 921 vv=(uu+1), Ngene
         cornum=0
         resnum=0
         DO 910 uu1=1,(Ngene-1)
           DO 911 vv1=(uu+1), Ngene
            DO 912 j=1, B
            IF (nullstat(uu1,vv1,j).GE.res(uu,vv)) resnum=resnum+1  
            IF (nullcorr(uu1,vv1,j).GE.corres(uu,vv)) cornum=cornum+1 
  912 continue         
  911 continue         
  910 continue         
         corpval(uu,vv)=real(cornum)/(Ngene*(Ngene-1)/2*B)
         newpval(uu,vv)=real(resnum)/(Ngene*(Ngene-1)/2*B)
  921 continue         
  900 continue         

       DO 950 uu=1, (Ngene-1)
          DO 952 vv=(uu+1), Ngene
          write (7,666) uu,vv,res(uu,vv),newpval(uu,vv),
     +                  corres(uu,vv),corpval(uu,vv)
  952  continue    
  950  continue    

  333  FORMAT (4X,'Gene1',4X,'Gene2',4X,'observed stat new',4X,
     +   'newpvalue',4X,'observed stat corr',4X,'corrpvalue')
  666  FORMAT (4X,I6,4X,I6,4X,F14.8,4X,F14.8,4X,F14.8,4X,F14.8) 


  111  FORMAT (4X,'Gene1',4X,'Gene2',4X,'new',4X,'corr',4X,'zstat') 
  222  FORMAT (4X,I6,4X, I6, 4X,F14.8,4X,F14.8,4X,F14.8) 
  444  FORMAT (4X,'Gene1',4X,'Gene2',9X,'B',9X,'nullstat',9X,'nullcorr')
  555  FORMAT (4X,I6,4X,I6,4X,I6,4X,F20.10,4X,F20.10)
  
       close(7)
       Close(8)       

       STOP
       END


c*********************************************************************       
      subroutine chivecfun(REP,T,qvec,chivec,nullmat,pivec)
c*********************************************************************       
c subroutine with input qvec; Output includes chivec, nullmat, pivec;
c 
c      
      INTEGER REP, T
      Double precision TMAT(4,4), indthat(REP,T-1,4,4)
      INTEGER N, tt, i, j,k, noncen
      Double precision temp(4),ssum,ssum1, ssum2
      Double precision  diff,TOTAL,ZERO
      PARAMETER (noncen=4,ZERO=0.0)           

      Double precision qvec(rep,T,4)
      Double precision config(T,4),betavec(rep,T,4)
      Double precision pivec(4)
      Double precision thatold(4,4),alphavec(rep,T,4)
      Double precision that(4,4),indconfig(rep,T,4),nullmat(4,4) 
      Double precision commart(4,4),nat(2,2),nbt(2,2),expnum(4,4),chivec
 
      pivec(1:4)=0.25
      that(1:4,1:4)=0.25
  
       diff=1.0
  900  if (diff .GE. 0.001) then
             thatold(1:4,1:4)=that(1:4,1:4)

          do 200 N=1, REP
              alphavec(N,1,1:4)=qvec(N,1,1:4)*pivec(1:4)

          do 220 tt=2, T   
c       call vecmat(4,4, thatold(1:4,1:4),alphavec(N,tt-1,1:4),temp)

         temp=matmul(alphavec(N,tt-1,1:4),thatold(1:4,1:4))

          alphavec(N,tt,1:4)=temp(1:4)*qvec(N,tt,1:4)

  220     continue

             betavec(N,T,1:4)=1.0

          do 250 tt=T-1, 1, -1
            temp(1:4)=qvec(N,tt+1,1:4)*betavec(N,tt+1,1:4)
          betavec(N,tt,1:4)=matmul(thatold(1:4,1:4),temp(1:4))

  250     continue

          do 270 k= 1, 4
            do 280 tt=1, T
              indconfig(N,tt,k)=alphavec(N,tt,k)*betavec(N,tt,k)/
     +   (dot_product(alphavec(N,tt,1:4),betavec(N,tt,1:4))+(1.0D-300))
  280       continue
  270     continue

          do 300 tt=1, T-1
            ssum=ZERO
            do 310 j=1, 4
              do 320 k=1, 4
                indthat(N,tt,j,k)=alphavec(N,tt,j)*thatold(j,k)
     +                       *qvec(N,(tt+1),k)*betavec(N,(tt+1),k)
                ssum=indthat(N,tt,j,k)+ssum
  320         continue
  310       continue
         indthat(N,tt,1:4,1:4)= indthat(N,tt,1:4,1:4)/(ssum+(1.0D-300))
  300     continue      
  200     continue

          do 400 tt=1, T
            do 410 j=1, 4
              ssum=ZERO
              do 420 i=1, rep
                ssum=indconfig(i,tt,j)+ssum
  420         continue
              config(tt,j)=ssum/rep
  410       continue
  400     continue

          that(1:4,1:4)=0.0    
          do 500 k=1, 4
            do 510 j=1, 4
              do 520 tt=1, T-1
                ssum1=ZERO
                do 530 i=1, rep
                  ssum1=indthat(i,tt,j,k)+ssum1
  530           continue
              that(j,k)=that(j,k)+ssum1/(config(tt,j)*REP+(1.0D-300))
  520         continue
  510       continue
  500     continue  
          that=that(1:4,1:4)/(T-1)

          ssum=ZERO
          do 600 k=1, 4
            do 610 j=1,4
                ssum=abs(that(j,k)-thatold(j,k))+ssum
  610       continue
  600     continue
          diff=ssum/16
          pivec(1:4)=config(1,1:4)  

          goto 900
          endif

       do 700 k=1, 4
          do 701 j=1, 4
             ssum=ZERO 
             do 702 tt=1, T-1
                do 703 N=1, REP
                   ssum=ssum+indthat(N,tt,j,k)
  703 continue
  702 continue
            commart(j,k)=ssum
  701 continue
  700 continue
         nat(1,1)=commart(1,1)+commart(1,2)+commart(2,1)+commart(2,2)
         nat(1,2)=commart(1,3)+commart(1,4)+commart(2,3)+commart(2,4)
         nat(2,1)=commart(3,1)+commart(3,2)+commart(4,1)+commart(4,2)
         nat(2,2)=commart(3,3)+commart(3,4)+commart(4,3)+commart(4,4)

         nbt(1,1)=commart(1,1)+commart(1,3)+commart(3,1)+commart(3,3)
         nbt(1,2)=commart(1,2)+commart(1,4)+commart(3,2)+commart(3,4)
         nbt(2,1)=commart(2,1)+commart(2,3)+commart(4,1)+commart(4,3)
         nbt(2,2)=commart(2,2)+commart(2,4)+commart(4,2)+commart(4,4)

         ssum=zero
         do 704 k=1, 2
            do 705 j=1, 2
               ssum=ssum+nat(j,k)
  705 continue
  704 continue

      call kronecker(nat, nbt, expnum)
      expnum(1:4,1:4)=expnum(1:4,1:4)/ssum

      nullmat=expnum/ssum
      chivec=ZERO
      do 706 j=1,4
        do 707 k=1, 4
           chivec=chivec+(commart(j,k)-expnum(j,k))**2/
     +            (expnum(j,k)+(1.0D-300))
  707   continue
  706 continue  

         RETURN
         END 


c*********************************************************************       
      Double precision function corr(n,x,y)
c*********************************************************************
c correlation between two vectors with the same length
c      
      integer n
      Double precision x(n), y(n)
      INTEGER i
      Double precision xsum,ysum, sum, xmean, ymean, xvar, yvar
      xsum=0.0
      ysum=0.0
      xvar=0.0
      yvar=0.0
      sum=0.0
      corr=0.0
      do 10 i=1, n
         xsum=xsum+x(i)
         ysum=ysum+y(i)
  10  continue
      xmean=xsum/n
      ymean=ysum/n  
      
      do 20 i=1, n
      xvar=xvar+(x(i)-xmean)**2
      yvar=yvar+(y(i)-ymean)**2
  20  continue
      
      do 30 i=1, n
      sum=sum+(x(i)-xmean)*(y(i)-ymean)
  30  continue
      corr=sum/(sqrt(xvar)*sqrt(yvar))

      return
      end


c********************************************************************* 
      Double precision function dnorm(x,Mean,SD)
c********************************************************************* 
c Density of Normal function with parameter mean and SD
c      
      Double precision x
      Double precision Mean,SD
      Double precision pi
      parameter(pi=3.1415926)
      dnorm = (1/(sqrt(2*pi)*SD))*EXP(-(x-mean)**2/(2*SD*SD))
      return
      end


c*********************************************************************
      subroutine kronecker(mat1,mat2,mat)
c*********************************************************************
c kronecker for 2x2 matrix
c      
      Double precision mat1(2,2), mat2(2,2), mat(4,4)
      INTEGER m, n, i, j
      do 50 i=1,4
         do 60 j=1,4
            mat(j,i)=0.0
  60  continue          
  50  continue
      do 10 i=1, 2
         do 20 j=1, 2
            do 30 m=1,2
               do 40 n=1,2
                 mat(m+2*(i-1),n+2*(j-1))=mat1(i,j)*mat2(m,n)
  40  continue
  30  continue
  20  continue
  10  continue
      RETURN
      END

c*********************************************************************
      subroutine normalization(m,n,x,y)   
c*********************************************************************
c normalization to each row of matrix sum to 1
c
      INTEGER m, n
      Double precision x(m,n), y(m,n)
      Double precision ssum
      INTEGER i,j
      Do 10 j=1, n
         Do 20 i=1, m
           y(m,n)=0.0
   20 continue
   10 continue         
      Do 30 i=1, m
         ssum=0.0
         Do 40 j=1, n       
            ssum=x(i,j)+ssum
   40 continue
         Do 50 j=1, n
            y(i,j)=x(i,j)/ssum
   50 continue
   30 continue
         RETURN
         END   


c*********************************************************************
      subroutine normalvec(m,x,y)   
c*********************************************************************
c normalization to each row of vec sum to 1
c
      INTEGER m
      Double precision x(m), y(m)
      Double precision ssum
      INTEGER j
           
      y=0.0
         ssum=0.0
         Do 40 j=1, m       
            ssum=x(j)+ssum
   40 continue
         Do 50 j=1, m
            y(j)=x(j)/ssum
   50 continue
 
         RETURN
         END   


C**********************************************************************C
C**********************************************************************C
      REAL FUNCTION snorm()
C**********************************************************************C
C                                                                      C
C                                                                      C
C     (STANDARD-)  N O R M A L  DISTRIBUTION                           C
C                                                                      C
C                                                                      C
C**********************************************************************C
C**********************************************************************C
C                                                                      C
C     FOR DETAILS SEE:                                                 C
C                                                                      C
C               AHRENS, J.H. AND DIETER, U.                            C
C               EXTENSIONS OF FORSYTHE'S METHOD FOR RANDOM             C
C               SAMPLING FROM THE NORMAL DISTRIBUTION.                 C
C               MATH. COMPUT., 27,124 (OCT. 1973), 927 - 937.          C
C                                                                      C
C     ALL STATEMENT NUMBERS CORRESPOND TO THE STEPS OF ALGORITHM 'FL'  C
C     (M=5) IN THE ABOVE PAPER     (SLIGHTLY MODIFIED IMPLEMENTATION)  C
C                                                                      C
C     Modified by Barry W. Brown, Feb 3, 1988 to use RANF instead of   C
C     SUNIF.  The argument IR thus goes away.                          C
C                                                                      C
C**********************************************************************C
C
      DIMENSION a(32),d(31),t(31),h(31)
C
C     THE DEFINITIONS OF THE CONSTANTS A(K), D(K), T(K) AND
C     H(K) ARE ACCORDING TO THE ABOVEMENTIONED ARTICLE
C
      DATA a/0.0,.3917609E-1,.7841241E-1,.1177699,.1573107,.1970991,
     +     .2372021,.2776904,.3186394,.3601299,.4022501,.4450965,
     +     .4887764,.5334097,.5791322,.6260990,.6744898,.7245144,
     +     .7764218,.8305109,.8871466,.9467818,1.009990,1.077516,
     +     1.150349,1.229859,1.318011,1.417797,1.534121,1.675940,
     +     1.862732,2.153875/
      DATA d/5*0.0,.2636843,.2425085,.2255674,.2116342,.1999243,
     +     .1899108,.1812252,.1736014,.1668419,.1607967,.1553497,
     +     .1504094,.1459026,.1417700,.1379632,.1344418,.1311722,
     +     .1281260,.1252791,.1226109,.1201036,.1177417,.1155119,
     +     .1134023,.1114027,.1095039/
      DATA t/.7673828E-3,.2306870E-2,.3860618E-2,.5438454E-2,
     +     .7050699E-2,.8708396E-2,.1042357E-1,.1220953E-1,.1408125E-1,
     +     .1605579E-1,.1815290E-1,.2039573E-1,.2281177E-1,.2543407E-1,
     +     .2830296E-1,.3146822E-1,.3499233E-1,.3895483E-1,.4345878E-1,
     +     .4864035E-1,.5468334E-1,.6184222E-1,.7047983E-1,.8113195E-1,
     +     .9462444E-1,.1123001,.1364980,.1716886,.2276241,.3304980,
     +     .5847031/
      DATA h/.3920617E-1,.3932705E-1,.3950999E-1,.3975703E-1,
     +     .4007093E-1,.4045533E-1,.4091481E-1,.4145507E-1,.4208311E-1,
     +     .4280748E-1,.4363863E-1,.4458932E-1,.4567523E-1,.4691571E-1,
     +     .4833487E-1,.4996298E-1,.5183859E-1,.5401138E-1,.5654656E-1,
     +     .5953130E-1,.6308489E-1,.6737503E-1,.7264544E-1,.7926471E-1,
     +     .8781922E-1,.9930398E-1,.1155599,.1404344,.1836142,.2790016,
     +     .7010474/
C
   10 u = ranf()
      s = 0.0
      IF (u.GT.0.5) s = 1.0
      u = u + u - s
   20 u = 32.0*u
      i = int(u)
      IF (i.EQ.32) i = 31
      IF (i.EQ.0) GO TO 100
C
C                                START CENTER
C
   30 ustar = u - float(i)
      aa = a(i)
   40 IF (ustar.LE.t(i)) GO TO 60
      w = (ustar-t(i))*h(i)
C
C                                EXIT   (BOTH CASES)
C
   50 y = aa + w
      snorm = y
      IF (s.EQ.1.0) snorm = -y
      RETURN
C
C                                CENTER CONTINUED
C
   60 u = ranf()
      w = u* (a(i+1)-aa)
      tt = (0.5*w+aa)*w
      GO TO 80

   70 tt = u
      ustar = ranf()
   80 IF (ustar.GT.tt) GO TO 50
   90 u = ranf()
      IF (ustar.GE.u) GO TO 70
      ustar = ranf()
      GO TO 40
C
C                                START TAIL
C
  100 i = 6
      aa = a(32)
      GO TO 120

  110 aa = aa + d(i)
      i = i + 1
  120 u = u + u
      IF (u.LT.1.0) GO TO 110
  130 u = u - 1.0
  140 w = u*d(i)
      tt = (0.5*w+aa)*w
      GO TO 160

  150 tt = u
  160 ustar = ranf()
      IF (ustar.GT.tt) GO TO 50
  170 u = ranf()
      IF (ustar.GE.u) GO TO 150
      u = ranf()
      GO TO 140

      END

C**********************************************************************
C**********************************************************************
      REAL FUNCTION ranf()
C**********************************************************************
C
C     REAL FUNCTION RANF()
C                RANDom number generator as a Function
C
C     Returns a random floating point number from a uniform distribution
C     over 0 - 1 (endpoints of this interval are not returned) using the
C     current generator
C
C     This is a transcription from Pascal to Fortran of routine
C     Uniform_01 from the paper
C
C     L'Ecuyer, P. and Cote, S. "Implementing a Random Number Package
C     with Splitting Facilities." ACM Transactions on Mathematical
C     Software, 17:98-111 (1991)
C
C**********************************************************************
C     .. External Functions ..
      INTEGER ignlgi
      EXTERNAL ignlgi
C     ..
C     .. Executable Statements ..
C
C     4.656613057E-10 is 1/M1  M1 is set in a data statement in IGNLGI
C      and is currently 2147483563. If M1 changes, change this also.
C
      ranf = ignlgi()*4.656613057E-10
      RETURN

      END


C**********************************************************************
C**********************************************************************
      INTEGER FUNCTION ignlgi()
C**********************************************************************
C
C     INTEGER FUNCTION IGNLGI()
C               GeNerate LarGe Integer
C
C     Returns a random integer following a uniform distribution over
C     (1, 2147483562) using the current generator.
C
C     This is a transcription from Pascal to Fortran of routine
C     Random from the paper
C
C     L'Ecuyer, P. and Cote, S. "Implementing a Random Number Package
C     with Splitting Facilities." ACM Transactions on Mathematical
C     Software, 17:98-111 (1991)
C
C**********************************************************************
C     .. Parameters ..
      INTEGER numg
      PARAMETER (numg=32)
C     ..
C     .. Scalars in Common ..
      INTEGER a1,a1vw,a1w,a2,a2vw,a2w,m1,m2
C     ..
C     .. Arrays in Common ..
      INTEGER cg1(numg),cg2(numg),ig1(numg),ig2(numg),lg1(numg),
     +        lg2(numg)
      LOGICAL qanti(numg)
C     ..
C     .. Local Scalars ..
      INTEGER curntg,k,s1,s2,z
      LOGICAL qqssd
C     ..
C     .. External Functions ..
      LOGICAL qrgnin
      EXTERNAL qrgnin
C     ..
C     .. External Subroutines ..
      EXTERNAL getcgn,inrgcm,rgnqsd,setall
C     ..
C     .. Common blocks ..
      COMMON /globe/m1,m2,a1,a2,a1w,a2w,a1vw,a2vw,ig1,ig2,lg1,lg2,cg1,
     +       cg2,qanti
C     ..
C     .. Save statement ..
      SAVE /globe/
C     ..
C     .. Executable Statements ..
C
C     IF THE RANDOM NUMBER PACKAGE HAS NOT BEEN INITIALIZED YET, DO SO.
C     IT CAN BE INITIALIZED IN ONE OF TWO WAYS : 1) THE FIRST CALL TO
C     THIS ROUTINE  2) A CALL TO SETALL.
C
      IF (.NOT. (qrgnin())) CALL inrgcm()
      CALL rgnqsd(qqssd)
      IF (.NOT. (qqssd)) CALL setall(1234567890,123456789)
C
C     Get Current Generator
C
      CALL getcgn(curntg)
      s1 = cg1(curntg)
      s2 = cg2(curntg)
      k = s1/53668
      s1 = a1* (s1-k*53668) - k*12211
      IF (s1.LT.0) s1 = s1 + m1
      k = s2/52774
      s2 = a2* (s2-k*52774) - k*3791
      IF (s2.LT.0) s2 = s2 + m2
      cg1(curntg) = s1
      cg2(curntg) = s2
      z = s1 - s2
      IF (z.LT.1) z = z + m1 - 1
      IF (qanti(curntg)) z = m1 - z
      ignlgi = z
      RETURN

      END

C**********************************************************************
C**********************************************************************
      LOGICAL FUNCTION qrgnin()
C**********************************************************************
C
C     LOGICAL FUNCTION QRGNIN()
C               Q Random GeNerators INitialized?
C
C     A trivial routine to determine whether or not the random
C     number generator has been initialized.  Returns .TRUE. if
C     it has, else .FALSE.
C
C**********************************************************************
C     .. Scalar Arguments ..
      LOGICAL qvalue
C     ..
C     .. Local Scalars ..
      LOGICAL qinit
C     ..
C     .. Entry Points ..
      LOGICAL qrgnsn
C     ..
C     .. Save statement ..
      SAVE qinit
C     ..
C     .. Data statements ..
      DATA qinit/.FALSE./
C     ..
C     .. Executable Statements ..
      qrgnin = qinit
      RETURN

C**********************************************************************
C**********************************************************************
      ENTRY qrgnsn(qvalue)
C**********************************************************************
C
C     LOGICAL FUNCTION QRGNSN( QVALUE )
C               Q Random GeNerators Set whether iNitialized
C
C     Sets state of whether random number generator is initialized
C     to QVALUE.
C
C     This routine is actually an entry in QRGNIN, hence it is a
C     logical function.  It returns the (meaningless) value .TRUE.
C
C**********************************************************************
      qinit = qvalue
      qrgnsn = .TRUE.
      RETURN

      END

C**********************************************************************
C**********************************************************************
      SUBROUTINE getcgn(g)
      INTEGER g
C**********************************************************************
C
C      SUBROUTINE GETCGN(G)
C                         Get GeNerator
C
C     Returns in G the number of the current random number generator
C
C
C                              Arguments
C
C
C     G <-- Number of the current random number generator (1..32)
C                    INTEGER G
C
C**********************************************************************
C
      INTEGER curntg,numg
      SAVE curntg
      PARAMETER (numg=32)
      DATA curntg/1/
C
      g = curntg
      RETURN

C**********************************************************************
C**********************************************************************
      ENTRY setcgn(g)
C**********************************************************************
C
C     SUBROUTINE SETCGN( G )
C                      Set GeNerator
C
C     Sets  the  current  generator to G.    All references to a generat
C     are to the current generator.
C
C
C                              Arguments
C
C
C     G --> Number of the current random number generator (1..32)
C                    INTEGER G
C
C**********************************************************************
C
C     Abort if generator number out of range
C
      IF (.NOT. (g.LT.0.OR.g.GT.numg)) GO TO 10
      WRITE (*,*) ' Generator number out of range in SETCGN:',
     +  ' Legal range is 1 to ',numg,' -- ABORT!'
      STOP ' Generator number out of range in SETCGN'

   10 curntg = g
      RETURN

      END

C**********************************************************************
C**********************************************************************
      SUBROUTINE inrgcm()
C**********************************************************************
C
C     SUBROUTINE INRGCM()
C          INitialize Random number Generator CoMmon
C
C
C                              Function
C
C
C     Initializes common area  for random number  generator.  This saves
C     the  nuisance  of  a  BLOCK DATA  routine  and the  difficulty  of
C     assuring that the routine is loaded with the other routines.
C
C**********************************************************************
C     .. Parameters ..
      INTEGER numg
      PARAMETER (numg=32)
C     ..
C     .. Scalars in Common ..
      INTEGER a1,a1vw,a1w,a2,a2vw,a2w,m1,m2
C     ..
C     .. Arrays in Common ..
      INTEGER cg1(numg),cg2(numg),ig1(numg),ig2(numg),lg1(numg),
     +        lg2(numg)
      LOGICAL qanti(numg)
C     ..
C     .. Local Scalars ..
      INTEGER i
      LOGICAL qdum
C     ..
C     .. External Functions ..
      LOGICAL qrgnsn
      EXTERNAL qrgnsn
C     ..
C     .. Common blocks ..
      COMMON /globe/m1,m2,a1,a2,a1w,a2w,a1vw,a2vw,ig1,ig2,lg1,lg2,cg1,
     +       cg2,qanti
C     ..
C     .. Save statement ..
      SAVE /globe/
C     ..
C     .. Executable Statements ..
C     V=20;                            W=30;
C
C     A1W = MOD(A1**(2**W),M1)         A2W = MOD(A2**(2**W),M2)
C     A1VW = MOD(A1**(2**(V+W)),M1)    A2VW = MOD(A2**(2**(V+W)),M2)
C
C   If V or W is changed A1W, A2W, A1VW, and A2VW need to be recomputed.
C    An efficient way to precompute a**(2*j) MOD m is to start with
C    a and square it j times modulo m using the function MLTMOD.
C
      m1 = 2147483563
      m2 = 2147483399
      a1 = 40014
      a2 = 40692
      a1w = 1033780774
      a2w = 1494757890
      a1vw = 2082007225
      a2vw = 784306273
      DO 10,i = 1,numg
          qanti(i) = .FALSE.
   10 CONTINUE
C
C     Tell the world that common has been initialized
C
      qdum = qrgnsn(.TRUE.)
      RETURN

      END

C**********************************************************************
C**********************************************************************
      SUBROUTINE setall(iseed1,iseed2)
C**********************************************************************
C
C      SUBROUTINE SETALL(ISEED1,ISEED2)
C               SET ALL random number generators
C
C     Sets the initial seed of generator 1 to ISEED1 and ISEED2. The
C     initial seeds of the other generators are set accordingly, and
C     all generators states are set to these seeds.
C
C     This is a transcription from Pascal to Fortran of routine
C     Set_Initial_Seed from the paper
C
C     L'Ecuyer, P. and Cote, S. "Implementing a Random Number Package
C     with Splitting Facilities." ACM Transactions on Mathematical
C     Software, 17:98-111 (1991)
C
C**********************************************************************
C     .. Parameters ..
      INTEGER numg
      PARAMETER (numg=32)
C     ..
C     .. Scalar Arguments ..
      INTEGER iseed1,iseed2
      LOGICAL qssd
C     ..
C     .. Scalars in Common ..
      INTEGER a1,a1vw,a1w,a2,a2vw,a2w,m1,m2
C     ..
C     .. Arrays in Common ..
      INTEGER cg1(numg),cg2(numg),ig1(numg),ig2(numg),lg1(numg),
     +        lg2(numg)
      LOGICAL qanti(numg)
C     ..
C     .. Local Scalars ..
      INTEGER g,ocgn
      LOGICAL qqssd
C     ..
C     .. External Functions ..
      INTEGER mltmod
      LOGICAL qrgnin
      EXTERNAL mltmod,qrgnin
C     ..
C     .. External Subroutines ..
      EXTERNAL getcgn,initgn,inrgcm,setcgn
C     ..
C     .. Common blocks ..
      COMMON /globe/m1,m2,a1,a2,a1w,a2w,a1vw,a2vw,ig1,ig2,lg1,lg2,cg1,
     +       cg2,qanti
C     ..
C     .. Save statement ..
      SAVE /globe/,qqssd
C     ..
C     .. Data statements ..
      DATA qqssd/.FALSE./
C     ..
C     .. Executable Statements ..
C
C     TELL IGNLGI, THE ACTUAL NUMBER GENERATOR, THAT THIS ROUTINE
C      HAS BEEN CALLED.
C
      qqssd = .TRUE.
      CALL getcgn(ocgn)
C
C     Initialize Common Block if Necessary
C
      IF (.NOT. (qrgnin())) CALL inrgcm()
      ig1(1) = iseed1
      ig2(1) = iseed2
      CALL initgn(-1)
      DO 10,g = 2,numg
          ig1(g) = mltmod(a1vw,ig1(g-1),m1)
          ig2(g) = mltmod(a2vw,ig2(g-1),m2)
          CALL setcgn(g)
          CALL initgn(-1)
   10 CONTINUE
      CALL setcgn(ocgn)
      RETURN

      ENTRY rgnqsd(qssd)
C**********************************************************************
C
C     SUBROUTINE RGNQSD
C                    Random Number Generator Query SeeD set?
C
C     Returns (LOGICAL) QSSD as .TRUE. if SETALL has been invoked,
C     otherwise returns .FALSE.
C
C**********************************************************************
      qssd = qqssd
      RETURN
      END


C**********************************************************************
C**********************************************************************
      INTEGER FUNCTION mltmod(a,s,m)
C**********************************************************************
C
C     INTEGER FUNCTION MLTMOD(A,S,M)
C
C                    Returns (A*S) MOD M
C
C     This is a transcription from Pascal to Fortran of routine
C     MULtMod_Decompos from the paper
C
C     L'Ecuyer, P. and Cote, S. "Implementing a Random Number Package
C     with Splitting Facilities." ACM Transactions on Mathematical
C     Software, 17:98-111 (1991)
C
C
C                              Arguments
C
C
C     A, S, M  -->
C                         INTEGER A,S,M
C
C**********************************************************************
C     .. Parameters ..
      INTEGER h
      PARAMETER (h=32768)
C     ..
C     .. Scalar Arguments ..
      INTEGER a,m,s
C     ..
C     .. Local Scalars ..
      INTEGER a0,a1,k,p,q,qh,rh
C     ..
C     .. Executable Statements ..
C
C     H = 2**((b-2)/2) where b = 32 because we are using a 32 bit
C      machine. On a different machine recompute H
C
      IF (.NOT. (a.LE.0.OR.a.GE.m.OR.s.LE.0.OR.s.GE.m)) GO TO 10
      WRITE (*,*) ' A, M, S out of order in MLTMOD - ABORT!'
      WRITE (*,*) ' A = ',a,' S = ',s,' M = ',m
      WRITE (*,*) ' MLTMOD requires: 0 < A < M; 0 < S < M'
      STOP ' A, M, S out of order in MLTMOD - ABORT!'

   10 IF (.NOT. (a.LT.h)) GO TO 20
      a0 = a
      p = 0
      GO TO 120

   20 a1 = a/h
      a0 = a - h*a1
      qh = m/h
      rh = m - h*qh
      IF (.NOT. (a1.GE.h)) GO TO 50
      a1 = a1 - h
      k = s/qh
      p = h* (s-k*qh) - k*rh
   30 IF (.NOT. (p.LT.0)) GO TO 40
      p = p + m
      GO TO 30

   40 GO TO 60

   50 p = 0
C
C     P = (A2*S*H)MOD M
C
   60 IF (.NOT. (a1.NE.0)) GO TO 90
      q = m/a1
      k = s/q
      p = p - k* (m-a1*q)
      IF (p.GT.0) p = p - m
      p = p + a1* (s-k*q)
   70 IF (.NOT. (p.LT.0)) GO TO 80
      p = p + m
      GO TO 70

   80 CONTINUE
   90 k = p/qh
C
C     P = ((A2*H + A1)*S)MOD M
C
      p = h* (p-k*qh) - k*rh
  100 IF (.NOT. (p.LT.0)) GO TO 110
      p = p + m
      GO TO 100

  110 CONTINUE
  120 IF (.NOT. (a0.NE.0)) GO TO 150
C
C     P = ((A2*H + A1)*H*S)MOD M
C
      q = m/a0
      k = s/q
      p = p - k* (m-a0*q)
      IF (p.GT.0) p = p - m
      p = p + a0* (s-k*q)
  130 IF (.NOT. (p.LT.0)) GO TO 140
      p = p + m
      GO TO 130

  140 CONTINUE
  150 mltmod = p
C
      RETURN
      END


C**********************************************************************
C**********************************************************************
      SUBROUTINE initgn(isdtyp)
C**********************************************************************
C
C     SUBROUTINE INITGN(ISDTYP)
C          INIT-ialize current G-e-N-erator
C
C     Reinitializes the state of the current generator
C
C     This is a transcription from Pascal to Fortran of routine
C     Init_Generator from the paper
C
C     L'Ecuyer, P. and Cote, S. "Implementing a Random Number Package
C     with Splitting Facilities." ACM Transactions on Mathematical
C     Software, 17:98-111 (1991)
C
C
C                              Arguments
C
C
C     ISDTYP -> The state to which the generator is to be set
C
C          ISDTYP = -1  => sets the seeds to their initial value
C          ISDTYP =  0  => sets the seeds to the first value of
C                          the current block
C          ISDTYP =  1  => sets the seeds to the first value of
C                          the next block
C
C                                   INTEGER ISDTYP
C
C**********************************************************************
C     .. Parameters ..
      INTEGER numg
      PARAMETER (numg=32)
C     ..
C     .. Scalar Arguments ..
      INTEGER isdtyp
C     ..
C     .. Scalars in Common ..
      INTEGER a1,a1vw,a1w,a2,a2vw,a2w,m1,m2
C     ..
C     .. Arrays in Common ..
      INTEGER cg1(numg),cg2(numg),ig1(numg),ig2(numg),lg1(numg),
     +        lg2(numg)
      LOGICAL qanti(numg)
C     ..
C     .. Local Scalars ..
      INTEGER g
C     ..
C     .. External Functions ..
      LOGICAL qrgnin
      INTEGER mltmod
      EXTERNAL qrgnin,mltmod
C     ..
C     .. External Subroutines ..
      EXTERNAL getcgn
C     ..
C     .. Common blocks ..
      COMMON /globe/m1,m2,a1,a2,a1w,a2w,a1vw,a2vw,ig1,ig2,lg1,lg2,cg1,
     +       cg2,qanti
C     ..
C     .. Save statement ..
      SAVE /globe/
C     ..
C     .. Executable Statements ..
C     Abort unless random number generator initialized
      IF (qrgnin()) GO TO 10
      WRITE (*,*) ' INITGN called before random number generator ',
     +  ' initialized -- abort!'
      STOP ' INITGN called before random number generator initialized'

   10 CALL getcgn(g)
      IF ((-1).NE. (isdtyp)) GO TO 20
      lg1(g) = ig1(g)
      lg2(g) = ig2(g)
      GO TO 50

   20 IF ((0).NE. (isdtyp)) GO TO 30
      CONTINUE
      GO TO 50
C     do nothing
   30 IF ((1).NE. (isdtyp)) GO TO 40
      lg1(g) = mltmod(a1w,lg1(g),m1)
      lg2(g) = mltmod(a2w,lg2(g),m2)
      GO TO 50

   40 STOP 'ISDTYP NOT IN RANGE'

   50 cg1(g) = lg1(g)
      cg2(g) = lg2(g)
      RETURN
      END

C*********************************************************************
c********************************************************************* 
      subroutine MTN (m,P,IR,dd)
c********************************************************************* 
c Multinomial function
c      
       Integer m,dd
       REAL P(dd),x, Plower, Pupper
       INTEGER i,IR(dd)

       CALL RANDOM_NUMBER(x) 
       IR=0
       Plower=0.
       Pupper=P(1)
       IF ((x.GT.Plower) .AND. (x.LE.Pupper))  IR(1)=1
 
       DO i=2, dd
       Plower=Pupper
       Pupper=P(i)+Plower  
       IF ((x.GT.Plower) .AND. (x.LE.Pupper))    IR(i)=1
       ENDDO 

       return
       end      
